import pygtk
pygtk.require('2.0')
import gtk, gobject, glib, cairo, pango

import erapp
from sheet import *
import erutils
import sys

#gtk.pygtk_version
#gtk.gtk_version
"""
Implemented:
- Select cell and range of cells
- Modify 'active' cell
- Simple formulas ()+-*/
- absolute and relative references to cells in formulas =A1 =[-1,-2]
- Copy source-range to destination-range
- Load a csv-file
- Save modifications as a csv-file

Todo:
- More tests on DR!!!
- Translate [A1]-form to [-2,0]-form dependent on the cell where formula is entered
- Handle country dependency of csv files
- Refresh should be made less intrusive
- Start with file from CTB
"""

APP_NAME = "Calc"
APP_VERSION = "0.7"

class CommandArea(gtk.HBox):
    def __init__(self):
        gtk.HBox.__init__(self)
        self.homogeneous = False
        self.spacing = 0

        # HOME
        self.btn_home = gtk.Button()
        self.btn_home.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_HOME, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_home.set_image(image)
        self.btn_home.connect("clicked", lambda w: app.go_home())
        self.pack_start(self.btn_home, False, False, 0)

        # LEFT
        self.btn_left = gtk.Button()
        self.btn_left.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_GO_BACK, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_left.set_image(image)
        self.btn_left.connect("clicked", lambda w: app.go_left(1))
        self.pack_start(self.btn_left, False, False, 0)

        # RIGHT
        self.btn_right = gtk.Button()
        self.btn_right.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_GO_FORWARD, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_right.set_image(image)
        self.btn_right.connect("clicked", lambda w: app.go_right(1))
        self.pack_start(self.btn_right, False, False, 0)

        # DOWN
        self.btn_down = gtk.Button()
        self.btn_down.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_GO_DOWN, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_down.set_image(image)
        self.btn_down.connect("clicked", lambda w: app.go_down(1))
        self.pack_start(self.btn_down, False, False, 0)

        # UP
        self.btn_up = gtk.Button()
        self.btn_up.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_GO_UP, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_up.set_image(image)
        self.btn_up.connect("clicked", lambda w: app.go_up(1))
        self.pack_start(self.btn_up, False, False, 0)

        # LOAD
        self.btn_load = gtk.Button()
        self.btn_load.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_OPEN, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_load.set_image(image)
        self.btn_load.connect("clicked", lambda w: app.do_load())
        self.pack_start(self.btn_load, False, False, 0)

        # STORE
        self.btn_store = gtk.Button()
        self.btn_store.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_SAVE, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_store.set_image(image)
        self.btn_store.connect("clicked", lambda w: app.do_store())
        self.pack_start(self.btn_store, False, False, 0)

        # COPY
        self.btn_copy = gtk.Button()
        self.btn_copy.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_COPY, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_copy.set_image(image)
        self.btn_copy.connect("clicked", lambda w: app.do_copy())
        self.pack_start(self.btn_copy, False, False, 0)

        # CLOSE
        self.btn_close = gtk.Button()
        self.btn_close.set_relief(gtk.RELIEF_NONE)
        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_STOP, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_close.set_image(image)
        self.btn_close.connect("clicked", lambda w: app.do_close())
        self.pack_end(self.btn_close, False, False, 0)

class EditArea(gtk.HBox):
    def __init__(self):
        gtk.HBox.__init__(self)
        self.homogeneous = False
        self.spacing = 0
        self.advance_mode_down = True
        
        # label
        self.label = gtk.Label()
        self.label.set_text('A1')
        self.pack_start(self.label, False, False, 10)
        
        # entry
        self.entry = gtk.Entry()
        self.entry.connect("activate", self.enter_callback, self.entry)
        self.entry.set_text('')
        self.entry.set_editable(False)
        self.entry.set_size_request(300, -1)
        self.pack_start(self.entry, False, False, 10)

        # advance direction button
        self.btn_dir = gtk.Button()
        self.btn_dir.set_relief(gtk.RELIEF_NONE)
        self.image_down = gtk.Image()
        self.image_down.set_from_stock(gtk.STOCK_GO_DOWN, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.image_right = gtk.Image()
        self.image_right.set_from_stock(gtk.STOCK_GO_FORWARD, gtk.ICON_SIZE_LARGE_TOOLBAR)
        self.btn_dir.set_image(self.image_down)
        self.btn_dir.connect("clicked", lambda w: self.toggle_advance())
        self.pack_start(self.btn_dir, False, False, 0)

    def toggle_advance(self):
        self.advance_mode_down = not self.advance_mode_down
        if self.advance_mode_down:
            self.btn_dir.set_image(self.image_down)
        else:
            self.btn_dir.set_image(self.image_right)
        app.set_advance_mode(self.advance_mode_down)
        
    def select_entry(self):
        """make the entry active"""
        self.entry.grab_focus()
        pass

    def set_value(self, cell, text, mode):
        self.entry.set_editable(mode)
        self.label.set_text(cell)
        self.entry.set_text(text)
        self.entry.select_region(0, len(text))

    def enter_callback(self, widget, entry):
        entry_text = self.entry.get_text()
        print "Entry contents: %s\n" % entry_text
        app.set_current(entry_text)


class CellArea(gtk.DrawingArea):
    # Signal handlers
    __gsignals__ = { "configure-event": "override",
                     "expose-event": "override",
                     "button-press-event": "override",
                     "button-release-event": "override"}

    def __init__(self):
        gtk.DrawingArea.__init__(self)
        self.set_double_buffered(False)
        self.set_app_paintable(True)
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK)
        self.col_width_list = None
        self.row_height_list = None
        self.start_x = 0
        self.start_y = 0
        self.gc = None
        self.bitmap = None
        # Need to set the size for Simulation
        self.set_size_request(700, 900)
        self.selected = [0,0]
        self.start_range = [0,0]
        self.end_range = [0,0]
        self.copy_range = [0,0,0,0]
        self.edit_mode = False
        self.copy_mode = False

    def do_configure_event(self, ev):
        self.width = ev.width
        self.height = ev.height
        allocation = self.get_allocation()
        print 'allocation (%d %d)' % (allocation.width, allocation.height)
        if (allocation.width > 0) and (allocation.height > 0):
            self.bitmap = gtk.gdk.Pixmap(self.window,
                                         allocation.width,
                                         allocation.height,
                                         -1)
            self.gc = gtk.gdk.GC(self.bitmap)
            # clear background
            self.draw_full()
        return True

    def do_expose_event(self, ev):
        #print 'on_expose (%d %d %d %d)' % (ev.area.x, ev.area.y, ev.area.width, ev.area.height)
        self.window.draw_drawable(self.gc,
                                  self.bitmap,
                                  ev.area.x, ev.area.y,
                                  ev.area.x, ev.area.y,
                                  ev.area.width, ev.area.height)
        return True

    def set_copy_mode(self, mode):
        self.copy_mode = mode
        
    def draw_full(self, mode='full'):
        # get sheet
        sheet = app.get_sheet(0)
        # draw the headers
        ## width of left size header and height of top header
        header_height = sheet.get_col_properties().find_property('header-height')
        self.header_height = header_height
        header_width = sheet.get_row_properties().find_property('header-width')
        self.header_width = header_width
        
        ## get column sizes
        col_width_list = sheet.get_col_widths(header_width, self.width)
        self.nr_cols = len(col_width_list)
        if self.col_width_list != None:
            del self.col_width_list
        self.col_width_list = col_width_list
        
        ## get row sizes
        row_height_list = sheet.get_row_heights(header_height, self.height) 
        self.nr_rows = len(row_height_list)
        if self.row_height_list != None:
            del self.row_height_list
        self.row_height_list = row_height_list
        #print "cols=%d rows=%d" %(self.nr_cols, self.nr_rows)
        
        # and draw them
        self.draw_background()
        self.draw_selection(header_width, header_height,
                            col_width_list, row_height_list)

        self.draw_col_header(header_width, 0,
                             self.width-header_width, header_height,
                             col_width_list, self.start_x)
        self.draw_row_header(0, header_height,
                             header_width, self.height-header_height,
                             row_height_list, self.start_y)
        self.draw_cell_borders(header_width, header_height,
                               self.width-header_width, self.height-header_height,
                               col_width_list, row_height_list)
        self.draw_cells(header_width, header_height,
                        self.width-header_width, self.height-header_height,
                        col_width_list, row_height_list, sheet,
                        self.start_x, self.start_y)

        #print 'draw_full'
            
        app.app_win.status_area.update() # label trick

    def sort_range(self):
        # make sure that start_range is upper_left and end_range is lower_right
        if self.start_range[0] > self.end_range[0]:
            self.start_range[0], self.end_range[0] = self.end_range[0], self.start_range[0]
        if self.start_range[1] > self.end_range[1]:
            self.start_range[1], self.end_range[1] = self.end_range[1], self.start_range[1]
            
    def calc_position(self, ev):
        x = self.header_width
        col = -1
        if x <= ev.x:
            col = 0
            while col < self.start_x:
                col += 1
            for col_w in self.col_width_list:
                if x + col_w > ev.x:
                    break
                x += col_w
                col += 1

        y = self.header_height
        row = -1
        if y <= ev.y:
            row = 0
            while row < self.start_y:
                row += 1
            for row_h in self.row_height_list:
                if y + row_h > ev.y:
                    break
                y += row_h
                row += 1

        return col, row

    def handle_select(self, col, row, force=False):
        if ((self.selected == [col, row]) and (self.edit_mode == False)) or force:
            # twice equal cell selected, go to edit mode
            self.edit_mode = True
            # clear range
            self.start_range = [-1,-1]
            self.end_range = [-1,-1]
        else:
            self.edit_mode = False
            
        self.selected = [col, row]
        #print "Range (%d,%d)-(%d,%d)" % (self.start_range[0],
        #                                 self.start_range[1],
        #                                 self.end_range[0],
        #                                 self.end_range[1])
        #print "Selected (%d,%d)" % (self.selected[0],
        #                            self.selected[1])

        self.draw_full()
        app.select_cell(self.selected[0], self.selected[1], self.edit_mode)
        self.queue_draw()
        
        
    def do_button_press_event(self, ev):
        #print "Position: %dx%d" % (ev.x, ev.y)

        col, row = self.calc_position(ev)
        self.start_range = [col, row]
        #print "Start Cell (%d,%d)" % (col, row)
        return True
    
    def do_button_release_event(self, ev):
        #print "Position: %dx%d" % (ev.x, ev.y)

        col, row = self.calc_position(ev)
        self.end_range = [col, row]
        self.sort_range()
        self.handle_select(col, row)

        if self.copy_mode:
            sheet = app.get_sheet(0)
            # copy the old range over to the new range
            src_x_range = self.copy_range[2] - self.copy_range[0] + 1
            src_y_range = self.copy_range[3] - self.copy_range[1] + 1
            dst_x_range = self.end_range[0] - self.start_range[0] + 1
            dst_y_range = self.end_range[1] - self.start_range[1] + 1
            #print 'src_range:', src_x_range, src_y_range
            #print 'dst_range:', dst_x_range, dst_y_range
            # if either is zero then dst range is too small
            for row in range(dst_y_range):
                for col in range(dst_x_range):
                    src_col = self.copy_range[0] + (col%src_x_range)
                    src_row = self.copy_range[1] + (row%src_y_range)
                    content = sheet.get_cell_content(src_col, src_row)
                    #print 'copy from :', src_col, src_row, content
                    dst_col = self.start_range[0] + col
                    dst_row = self.start_range[1] + row
                    sheet.set_cell_content(dst_col, dst_row, content)
                    #print 'copy to :', dst_col, dst_row
            
            self.copy_mode = False
            # redraw
            self.draw_full()
            self.queue_draw() # refresh

        # remember 'old' range for copy
        self.copy_range = [self.start_range[0], self.start_range[1], self.end_range[0], self.end_range[1]]
        return True
    
    def draw_background(self):
        # clear background
        allocation = self.get_allocation()
        white = gtk.gdk.Color(0xffff, 0xffff, 0xffff, 0)
        self.gc.set_rgb_fg_color(white)
        #print 'draw_background (%d %d)' % (allocation.width, allocation.height)
        self.bitmap.draw_rectangle(self.style.white_gc,
                                   True,
                                   0,0,
                                   allocation.width,
                                   allocation.height)

    def draw_selection(self, x, y, col_width_list, row_height_list):
        # check if in visible area
        if (self.start_range[0] - self.start_x) > len(col_width_list):
            return
        if (self.start_range[1] - self.start_y) > len(row_height_list):
            return
        # do selection area
        grey = gtk.gdk.Color(0xcccc, 0xcccc, 0xcccc, 12)
        self.gc.set_rgb_fg_color(grey)
        #print 'draw_selection;',self.start_range[0], self.start_range[1], self.end_range[0], self.end_range[1]
        if (self.start_range[0] >= 0) and (self.start_range[1] >= 0) and (self.end_range[0] >= 0) and (self.end_range[1] >= 0):
            col_nr = 0
            sum_width = x
            width = 0
            while col_nr < (self.start_range[0] - self.start_x):
                col_nr += 1
                sum_width += col_width_list[col_nr]
            while col_nr <= (self.end_range[0] - self.start_x):
                col_nr += 1
                width += col_width_list[col_nr]
            #print 'X:', sum_width, width
                
            row_nr = 0
            sum_height = y
            height = 0
            while row_nr < (self.start_range[1] - self.start_y):
                row_nr += 1
                sum_height += row_height_list[row_nr]
            while row_nr <= (self.end_range[1] - self.start_y):
                row_nr += 1
                height += row_height_list[row_nr]
            #print 'Y:', sum_height, height
            self.bitmap.draw_rectangle(self.gc,
                                       True,
                                       sum_width, sum_height,
                                       width,
                                       height)

        
        
    def draw_col_header(self, x, y, w, h, col_width_list, start):
        #print 'draw_col_header %d %d %d %d' %(x, y, w, h)
        # draw properties
        black = gtk.gdk.Color(0, 0, 0, 1)
        self.gc.set_rgb_fg_color(black)
        self.gc.set_line_attributes(2,
                                    gtk.gdk.LINE_SOLID,
                                    gtk.gdk.CAP_ROUND,
                                    gtk.gdk.JOIN_ROUND)
        font_str = "Sans %dpx" % (2*20/3)
        font_desc = pango.FontDescription(font_str)
        layout = self.create_pango_layout("A")
        layout.set_font_description(font_desc)
        
        #layout.set_width(pango.SCALE * self.allocation.width)
        #layout.set_spacing(pango.SCALE * 3)
        #layout.set_alignment(pango.ALIGN_CENTER)

        # determine number of columns that fit
        col = start
        col_names = "ABCDEFGHIJKLMNOPQRSTUVW"
        sum_width = 0
        for col_w in col_width_list:
            self.bitmap.draw_rectangle(self.gc,
                                       False,
                                       x + sum_width, y,
                                       col_w, h)
            layout.set_text(col_names[col])
            inkRect, (x1, y1, w1, h1) = layout.get_pixel_extents()
            #width, height = self.get_size_request()
            # position: center of cell: x + sum_width + col_w / 2, h/2
            ox = x + sum_width + col_w / 2 - (x1 + w1)/2
            oy = h/2 - (h1+y1)/2
            self.bitmap.draw_layout(self.gc,
                                    ox, oy,
                                    layout)
            sum_width += col_w
            col += 1
        #glib.unref(layout)
        #layout.unref()

    def draw_row_header(self, x, y, w, h, row_height_list, start):
        # draw properties
        black = gtk.gdk.Color(0, 0, 0, 1)
        self.gc.set_rgb_fg_color(black)
        self.gc.set_line_attributes(2,
                                    gtk.gdk.LINE_SOLID,
                                    gtk.gdk.CAP_ROUND,
                                    gtk.gdk.JOIN_ROUND)
        font_str = "Sans %dpx" % (2*20/3)
        font_desc = pango.FontDescription(font_str)
        layout = self.create_pango_layout("A")
        layout.set_font_description(font_desc)

        # determine number of rows that fit
        row = start
        sum_height = 0
        for row_h in row_height_list:
            self.bitmap.draw_rectangle(self.gc,
                                       False,
                                       x , y + sum_height,
                                       w, row_h)

            row_name = str(row+1)
            layout.set_text(row_name)
            inkRect, (x1, y1, w1, h1) = layout.get_pixel_extents()
            #width, height = self.get_size_request()
            # position: center of cell: x + w/2, y + sum_height + row_h/2
            ox = x + w/2 - (w1 + x1)/2
            oy = y + sum_height + row_h/2 - (h1 + y1)/2
            self.bitmap.draw_layout(self.gc,
                                    ox, oy,
                                    layout)
            sum_height += row_h
            row += 1
        #glib.unref(layout)
        #layout.unref()
            
    def draw_cell_borders(self, x, y, w, h, col_width_list, row_height_list):
        # draw properties
        grey = gtk.gdk.Color(0x8888, 0x8888, 0x8888, 7)
        self.gc.set_rgb_fg_color(grey)
        self.gc.set_line_attributes(1,
                                    gtk.gdk.LINE_SOLID,
                                    gtk.gdk.CAP_ROUND,
                                    gtk.gdk.JOIN_ROUND)
        
        # columns
        sum_width = x
        for col_w in col_width_list:
            sum_width += col_w
            self.bitmap.draw_line(self.gc, sum_width, y, sum_width, y+h)
        # rows
        sum_height = y
        for row_h in row_height_list:
            sum_height += row_h
            self.bitmap.draw_line(self.gc, x, sum_height, x+w, sum_height)

        #print 'Selection (%d,%d)' % (self.selected[0], self.selected[0])
        # Highlight selection
        if ((self.selected[0] - self.start_x) >= 0) and ((self.selected[1] - self.start_y) >= 0):
            # check if in visible area
            if (self.selected[0] - self.start_x) > len(col_width_list):
                return
            if (self.selected[1] - self.start_y) > len(row_height_list):
                return
            # only single-cell selection
            col_nr = 0
            sum_width = x
            while col_nr < (self.selected[0] - self.start_x):
                col_nr += 1
                sum_width += col_width_list[col_nr]
            width = col_width_list[col_nr]
            #print 'X:', sum_width, width
                
            row_nr = 0
            sum_height = y
            while row_nr < (self.selected[1] - self.start_y):
                row_nr += 1
                sum_height += row_height_list[row_nr + self.start_y]
            height = row_height_list[row_nr + self.start_y]
            #print 'Y:', sum_height, height

            black = gtk.gdk.Color(0, 0, 0, 1)
            self.gc.set_rgb_fg_color(black)
            self.gc.set_line_attributes(2,
                                        gtk.gdk.LINE_SOLID,
                                        gtk.gdk.CAP_ROUND,
                                        gtk.gdk.JOIN_ROUND)
            self.bitmap.draw_line(self.gc,
                                  sum_width, sum_height,
                                  sum_width + width, sum_height)
            self.bitmap.draw_line(self.gc,
                                  sum_width, sum_height,
                                  sum_width, sum_height + height)
            self.bitmap.draw_line(self.gc,
                                  sum_width, sum_height + height,
                                  sum_width + width, sum_height + height)
            self.bitmap.draw_line(self.gc,
                                  sum_width + width, sum_height,
                                  sum_width + width, sum_height + height)
                

    def draw_cells(self, x, y, w, h, col_width_list, row_height_list, sheet, start_x, start_y):
        # invalidate pre-calculated values
        sheet.invalidate()
        
        # draw properties
        black = gtk.gdk.Color(0, 0, 0, 1)
        self.gc.set_rgb_fg_color(black)
        self.gc.set_line_attributes(2,
                                    gtk.gdk.LINE_SOLID,
                                    gtk.gdk.CAP_ROUND,
                                    gtk.gdk.JOIN_ROUND)
        font_str = "Sans %dpx" % (2*20/3)
        font_desc = pango.FontDescription(font_str)
        layout = self.create_pango_layout("A")
        layout.set_font_description(font_desc)
        # determine number of rows that fit
        sum_height = y
        row_nr = start_y
        for row_h in row_height_list:
            sum_width = x
            col_nr = start_x
            for col_w in col_width_list:
                text = sheet.get_cell_value(col_nr, row_nr)
                if text != None:
                    layout.set_text(text)
                    inkRect, (x1, y1, w1, h1) = layout.get_pixel_extents()
                    #width, height = self.get_size_request()
                    # position: center of cell: x + w/2, y + sum_height + row_h/2
                    ox = sum_width + 1 #- (w1 + x1)/2
                    oy = sum_height + row_h/2 - (h1 + y1)/2
                    self.bitmap.draw_layout(self.gc,
                                            ox, oy,
                                            layout)
                sum_width += col_w
                col_nr += 1
            sum_height += row_h
            row_nr += 1
        #glib.unref(layout)
        #layout.unref()

    def do_left(self, step):
        self.start_x -= step
        if self.start_x < 0:
            self.start_x = 0
        self.update()

    def do_right(self, step):
        self.start_x += step
        if self.start_x > 20:
            # limit scrolling
            self.start_x = 20
        self.update()

    def do_up(self, step):
        self.start_y -= step
        if self.start_y < 0:
            self.start_y = 0
        self.update()

    def do_down(self, step):
        self.start_y += step
        if self.start_y > 1000:
            # limit scrolling
            self.start_y = 1000
        self.update()

    def do_goto(self, x, y):
        self.start_x = x
        self.start_y = y
        self.update()

    def update(self):
        #if not self.get_realized():
        #    print "Not yet realized"
        #    return
        #print "update"
        #app.refresh_start()
        self.draw_full()
        #app.refresh_end()
        self.queue_draw()

class StatusArea(gtk.HBox):
    def __init__(self):
        gtk.HBox.__init__(self)
        self.homogeneous = False
        self.spacing = 0
        self.set_size_request(-1, 20)
        self.hdr = gtk.Label("")
        #self.hdr.set_markup("<b><span size='xx-large'>Status Area</span></b>")
        self.hdr.set_text("Calc")
        self.pack_start(self.hdr, True, True, 10)

    def update(self):
        # the old fashion force update
        self.hdr.set_text("Calc")

class AppWindow(gtk.VBox):
    def __init__(self):
        gtk.VBox.__init__(self)
        self.homogeneous = False
        self.spacing = 0
        # Command Area
        self.command_area = CommandArea()
        self.pack_start(self.command_area, False, False, 0)
        # Edit Area
        self.edit_area = EditArea()
        self.pack_start(self.edit_area, False, False, 0)
        # Cell Area
        self.cell_area = CellArea()
        self.pack_start(self.cell_area, True, True, 0)
        # Status Area
        self.status_area = StatusArea()
        self.pack_start(self.status_area, False, False, 0)

def on_send_startup_complete():
    try:
        win = app.window.xid
    except:
        # in simulator on x86 this does not work, so use substitue
        win = 1
    app.ipc.send_startup_complete(win)
    return False

class AppCalc(erapp.DrApplication, gtk.Window):
    def __init__(self, filename=None):
        # create an empty sheet
        self.sheet = Sheet('Sheet1')
        self.edit_mode = False
        self.advance_mode_down = True
        self.filename = filename
        if filename != None:
            self.sheet.load_from_csv(filename)

    def draw_ui(self):
        gtk.Window.__init__(self)
        # display management
        self.erutils = erutils.ErUtils()
        self.have_control = False
        # create ui and dr specific stuff
        self.build_ui()
        self.build_dr()
        # announce that app is started (and can be added to task bar)
        glib.idle_add(on_send_startup_complete)

    def get_sheet(self, nr):
        return self.sheet

    def set_advance_mode(self, down):
        self.advance_mode_down = down

    def build_ui(self):
        self.set_title("calc")
        self.connect("destroy", self.on_quit)
        self.connect("key-press-event", self.do_key_pressed_event)
        self.connect("key-release-event", self.do_key_released_event)
        # Create content
        self.app_win = AppWindow()
        self.add(self.app_win)
        self.show_all()

    def on_quit(self, widget):
        print "Quiting"
        self.ipc.set_keyboard(False)
        gtk.main_quit()
        pass

    def do_key_pressed_event(self, widget, key):
        #print gtk.gdk.keyval_name(key.keyval)
        if self.edit_mode:
            return False
        else:
            return True

    def do_key_released_event(self, widget, ev):
        keyname = gtk.gdk.keyval_name(ev.keyval)
        #print "Key-up: %d-%d%s" % (ev.state, ev.keyval, keyname)
        if self.edit_mode:
            return False
        else:
            stepsize = 1
            if ev.state == gtk.gdk.SHIFT_MASK:
                stepsize *= 10
            if keyname == 'Left':
                self.go_left(stepsize)
            elif keyname == 'Right':
                self.go_right(stepsize)
            elif keyname == 'Up':
                self.go_up(stepsize)
            elif keyname == 'Down':
                self.go_down(stepsize)
            elif keyname == 'Page_Up':
                self.go_up(10*stepsize)
            elif keyname == 'Page_Down':
                self.go_down(10*stepsize)
            elif keyname == 'Home':
                self.go_home()
            elif keyname == 'End':
                print "End"
            elif keyname == 'BackSpace':
                print "BackSpace"
            # for simulation on PC
            elif keyname == 'a':
                self.do_about(APP_NAME, APP_VERSION)
            return True

    def build_dr(self):
        # For the name no spaces are allowed, only use lower case characters
        self.ipc = erapp.DrIpc("calc", "1.0", self);
        menu_manager = erapp.MenuManager(self.ipc);
        self.menu_manager = menu_manager
        # menu
        menu_main = erapp.Menu(menu_manager, "menumain", "Calc");
        self.menu_main = menu_main
        group_main = menu_main.addGroup("groupmain", "Main Buttons");

        group_main.addItem("close",       "Close",        "close");
        group_main.addItem("load",        "Load",         "open");
        group_main.addItem("about",       "About",        "info");
        menu_main.realise();

        # toolbar
        menu_manager.add_toolbar_item("calc_menumain", "calc_groupmain",  "close");
        menu_manager.add_toolbar_item("calc_menumain", "calc_groupmain",  "load");
        menu_manager.add_toolbar_item("calc_menumain", "calc_groupmain",  "about");

        # show keyboard
        # todo: keyboard gets removed when entrry-widget is deselected
        #       find a solution for that before enabling keyboard at startup
        #self.ipc.set_keyboard(True)
        menu_main.show();

    def onMenuClick(self, item, group, menu, state):
        # group in concatenation of ipc-name and group-name
        if group == "calc_groupmain":
            if item == "close":
                self.do_close()
            elif item == "load":
                self.do_load()
            elif item == "about":
                self.do_about(APP_NAME, APP_VERSION)

    def onFileOpen(self, filename):
        print 'onFileOpen'

    def onFileClose(self, filename):
        print 'onFileClose'

    def onWindowChange(self, xid, activated):
        if activated:
            # Need to change menu when we are activated again
            self.menu_main.show()

    def onPrepareUnmount(self, device):
        print 'onPrepareUnmount'

    def onUnmounted(self, device):
        print 'onUnmounted'

    def onMounted(self, device):
        print 'onMounted'

    def onPrepareHibernate(self):
        print 'onPrepareHibernate'

    def onChangeLocale(self, locale):
        print 'onChangeLocale'

    def onChangedOrientation(self, orientation):
        print 'onChangedOrientation'

    def getMainWindow(self):
        return self

    def go_left(self, step):
        self.app_win.cell_area.do_left(step)
        
    def go_right(self, step):
        self.app_win.cell_area.do_right(step)

    def go_up(self, step):
        self.app_win.cell_area.do_up(step)

    def go_down(self, step):
        self.app_win.cell_area.do_down(step)
        
    def go_home(self):
        self.app_win.cell_area.do_goto(0, 0)
        
    def do_close(self):
        self.on_quit(None)
        
    def do_load(self):
        filename = self.run_open_dialog()
        if filename != None:
            self.sheet.load_from_csv(filename)
        self.app_win.cell_area.update()

    def do_store(self):
        filename = self.run_store_dialog()
        if filename != None:
            self.sheet.save_to_csv(filename)

    def do_copy(self):
        # enable copy mode, next selection will lead to copy
        self.app_win.cell_area.set_copy_mode(True)
    
    def on_idle_display_yield(self):
        if self.have_control == True:
            self.erutils.display_update_return_control(4) # 1:DM_HINT_FULL, 4:DM_HINT_PARTIAL
            self.have_control = False
            print 'yield'

    def refresh_start(self):
        #if self.have_control == False:
        #    self.erutils.display_gain_control()
        #    self.have_control = True
        pass
            
    def refresh_end(self, mode=0):
        # helper function to force a refresh of the screen
        self.app_win.status_area.update() # label trick
        #if self.have_control == True:
        #    if mode == 0:
        #        # for drawing
        #        glib.idle_add(self.on_idle_display_yield)
        #    else:
        #        # extra time for dialogs to be removed
        #        glib.timeout_add(400, self.on_idle_display_yield)
        
    def dialog_default_size_changed_cb(self, widget, dialog):
        # this callback makes sure that the size of the dialog is big enough
        width = 700
        height = 700
        dialog.set_size_request(width, height)
        dialog.set_resizable(True)

    def dialog_response_requested_cb(self, widget, dialog):
        # This is used in xournal, but seems not to be required
        print "dialog_response_requested_cb"
    
    def dialog_file_activated_cb(self, widget, dialog):
        if dialog.activate_default():
            return
        # this should not be executed ...
        print  'dialog_file_activated_cb: no default'

    def do_about(self, name, version):
        dialog = gtk.Dialog("About", None,
                            gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_OK, gtk.RESPONSE_OK))
        vbox = dialog.vbox

        # Application Name
        label = gtk.Label(name)
        label.show()
        vbox.pack_start(label, False, False, 3)
        label.set_justify(gtk.JUSTIFY_CENTER)

        # Application Version
        ver_str = "\nVersion %s\n" % version
        label = gtk.Label(ver_str)
        label.show()
        vbox.pack_start(label, False, False, 3)
        label.set_justify(gtk.JUSTIFY_CENTER)

        # create some space in the dialog
        label = gtk.Label("                                      ")
        label.show()
        vbox.pack_start(label, False, False, 3)
        label.set_justify(gtk.JUSTIFY_CENTER)

        vbox.show()
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.run()
        dialog.destroy()
        
    def run_open_dialog(self):
        dialog = gtk.Dialog( "Open..", None,
                             gtk.DIALOG_DESTROY_WITH_PARENT,
                             (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                              gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_has_separator(False)
        dialog.set_border_width(5)
        dialog.vbox.set_spacing(2)
        dialog.action_area.set_border_width(5)
        dialog.set_default_response(gtk.RESPONSE_OK)

        #dialog.connect("response", self.dialog_response_cb)
        widget = gtk.FileChooserWidget()
        widget.connect("file-activated", lambda w: self.dialog_file_activated_cb(w, dialog))
        widget.connect("default-size-changed", lambda w: self.dialog_default_size_changed_cb(w, dialog))
        widget.connect("response-requested", lambda w: self.dialog_response_requested_cb(w, dialog))

        widget.set_border_width(5)
        dialog.vbox.pack_start(widget, True, True, 0)

        filter = gtk.FileFilter()
        filter.set_name("csv files")
        filter.add_pattern("*.csv")
        widget.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        widget.add_filter(filter)

        widget.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)

        widget.show()

        if dialog.run() != gtk.RESPONSE_OK:
            dialog.destroy()
            return None

        filename = widget.get_filename()
        dialog.destroy()

        return filename

    def run_store_dialog(self):
        dialog = gtk.Dialog( "Store..", None,
                             gtk.DIALOG_DESTROY_WITH_PARENT,
                             (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                              gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_has_separator(False)
        dialog.set_border_width(5)
        dialog.vbox.set_spacing(2)
        dialog.action_area.set_border_width(5)
        dialog.set_default_response(gtk.RESPONSE_OK)

        #dialog.connect("response", self.dialog_response_cb)
        widget = gtk.FileChooserWidget()
        widget.connect("file-activated", lambda w: self.dialog_file_activated_cb(w, dialog))
        widget.connect("default-size-changed", lambda w: self.dialog_default_size_changed_cb(w, dialog))
        widget.connect("response-requested", lambda w: self.dialog_response_requested_cb(w, dialog))

        widget.set_border_width(5)
        dialog.vbox.pack_start(widget, True, True, 0)

        widget.set_action(gtk.FILE_CHOOSER_ACTION_SAVE)
        #widget.set_filename(filename)
        #widget.set_current_name(basename(filename)
        widget.set_current_name('MyFile.csv')

        filter = gtk.FileFilter()
        filter.set_name("csv files")
        filter.add_pattern("*.csv")
        widget.add_filter(filter)
        filter = gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        widget.add_filter(filter)

        widget.show()

        if dialog.run() != gtk.RESPONSE_OK:
            dialog.destroy()
            return None

        filename = widget.get_filename()
        print 'store:', filename
        dialog.destroy()

        print 'store:', filename
        return filename

    def select_cell(self, col, row, edit_mode):
        if (col >= 0) and (row >= 0):
            # check if it is a valid cell
            value = self.sheet.get_cell_content(col, row)
            if value == None:
                value = ''

            self.edit_mode = edit_mode
            col_names = "ABCDEFGHIJKLMNOPQRSTUVW"
            name = ''
            if col > 26:
                name += col_names[col/26]
            name += col_names[col%26]
            name += str(row+1)
            app.app_win.edit_area.set_value(name, value, edit_mode)
            app.app_win.edit_area.select_entry()
            self.current_edit = [col, row]
        else:
            app.app_win.edit_area.set_value('--', '', False)
            self.current_edit = [-1, -1]

    def advance_cell(self, col, row):
        if self.advance_mode_down:
            return col, row+1
        else:
            return col+1, row
        
    def set_current(self, value):
        if (self.current_edit[0] >= 0) and (self.current_edit[1] >= 0):
            self.sheet.add_cell(self.current_edit[0], self.current_edit[1], value)
            col, row = self.advance_cell(self.current_edit[0], self.current_edit[1])
            self.app_win.cell_area.handle_select(col, row, True)
            #self.app_win.cell_area.update()

# on the DR this file is imported from a generic file called 'python'
# (to overcome some DR limitations). So we can not test for if __name__ == "__main__":
if len(sys.argv) > 1:
    print 'Found argument:', sys.argv[1]
    filename = sys.argv[1]
app = AppCalc(filename)
app.draw_ui()
gtk.main()
